@@ -14,11 +14,13 @@ from pathlib import Path
 from tempfile import TemporaryDirectory
 from threading import Lock
 from types import ModuleType
-from typing import Callable, Optional, TextIO, Tuple, Union
+from typing import Callable, Optional, TextIO, Tuple, TypeVar, Union, overload
 
 from autoregistry import Registry
 from pathspec import PathSpec
 from serial import SerialException
+from serial.tools.miniterm import Miniterm
+from typing_extensions import ParamSpec
 
 from ._minify import minify as minify_code
 from .exceptions import ConnectionLost, MaxHistoryLengthError
@@ -36,6 +38,9 @@ from .pyboard import Pyboard, PyboardError, PyboardException
 from .typing import BelayReturn, PathType
 from .webrepl import WebreplToSerial
 
+P = ParamSpec("P")
+R = TypeVar("R")
+
 
 class NotBelayResponse(Exception):
     """Parsed response wasn't for Belay."""
@@ -230,6 +235,10 @@ class MethodMetadata:
 class Device(Registry):
     """Belay interface into a micropython device.
 
+    Can be used as a context manager; calls ``self.close`` on exit.
+
+    Inherits from ``autoregistry.Registry`` for easy access to subclasses.
+
     Attributes
     ----------
     implementation: Implementation
@@ -648,14 +657,24 @@ class Device(Registry):
         self.close()
 
     def close(self) -> None:
-        """Close the connection to device."""
+        """Close the connection to device.
+
+        Automatically called on context manager exit.
+        """
         # Invoke all teardown executers prior to closing out connection.
+        if self._board is None:
+            # Has already been closed
+            return
+
         atexit.unregister(self.close)
 
+        self._board.cancel_running_program()
+
         for executer in _sort_executers(self._belay_teardown._belay_executers):
             executer()
 
-        return self._board.close()
+        self._board.close()
+        self._board = None
 
     def reconnect(self, attempts: Optional[int] = None) -> None:
         """Reconnect to the device and replay the command history.
@@ -684,11 +703,23 @@ class Device(Registry):
         for cmd in self._cmd_history:
             self(cmd, record=False)
 
+    @overload
     @staticmethod
-    def setup(f=None, *, autoinit=False, **kwargs) -> Callable:
-        """setup(f, *, minify=True, register=True, record=True)
+    def setup(f: Callable[P, R]) -> Callable[P, R]:
+        ...
 
-        Decorator that executes function's body in a global-context on-device when called.
+    @overload
+    @staticmethod
+    def setup(
+        *, autoinit=False, **kwargs
+    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
+        ...
+
+    @staticmethod
+    def setup(
+        f: Optional[Callable[P, R]] = None, autoinit=False, **kwargs
+    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
+        """Execute decorated function's body in a global-context on-device when called.
 
         Function arguments are also set in the global context.
 
@@ -714,7 +745,7 @@ class Device(Registry):
         """  # noqa: D400
         if f is None:
             return wraps_partial(Device.setup, autoinit=autoinit, **kwargs)  # type: ignore[reportGeneralTypeIssues]
-        if signature(f).parameters:
+        if signature(f).parameters and autoinit:
             raise ValueError(
                 f"Method {f} decorated with "
                 '"@Device.setup(autoinit=True)" '
@@ -726,12 +757,21 @@ class Device(Registry):
         )
         return f
 
+    @overload
     @staticmethod
-    def teardown(f=None, **kwargs) -> Callable:
-        """teardown(f, *, minify=True, register=True, record=True)
+    def teardown(f: Callable[P, R]) -> Callable[P, R]:
+        ...
 
-        Decorator that executes function's body in a global-context on-device
-        when ``device.close()`` is called.
+    @overload
+    @staticmethod
+    def teardown(**kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
+        ...
+
+    @staticmethod
+    def teardown(
+        f: Optional[Callable[P, R]] = None, **kwargs
+    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
+        """Executes decorated function's body in a global-context on-device when ``device.close()`` is called.
 
         Function arguments are also set in the global context.
 
@@ -762,11 +802,24 @@ class Device(Registry):
         f.__belay__ = MethodMetadata(executer=TeardownExecuter, kwargs=kwargs)
         return f
 
+    @overload
     @staticmethod
-    def task(f=None, **kwargs) -> Callable:
-        """task(f, *, minify=True, register=True, record=False)
+    def task(f: Callable[P, R]) -> Callable[P, R]:
+        ...
 
-        Decorator that send code to device that executes when decorated function is called on-host.
+    @overload
+    @staticmethod
+    def task(**kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
+        ...
+
+    @staticmethod
+    def task(
+        f: Optional[Callable[P, R]] = None, **kwargs
+    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
+        """Execute decorated function on-device.
+
+        Sends source code to device at decoration time.
+        Execution sends involves much smaller overhead.
 
         Can either be used as a staticmethod ``@Device.task`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.task`` for marking functions to a specific ``Device`` instance.
 
@@ -792,10 +845,21 @@ class Device(Registry):
         return f
 
     @staticmethod
-    def thread(f=None, **kwargs) -> Callable:
-        """thread(f, *, minify=True, register=True, record=True)
+    @overload
+    @staticmethod
+    def thread(f: Callable[P, R]) -> Callable[P, R]:
+        ...
 
-        Decorator that send code to device that spawns a thread when executed.
+    @overload
+    @staticmethod
+    def thread(**kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
+        ...
+
+    @staticmethod
+    def thread(
+        f: Optional[Callable[P, R]] = None, **kwargs
+    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
+        """Spawn on-device thread that executes decorated function.
 
         Can either be used as a staticmethod ``@Device.thread`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.thread`` for marking functions to a specific ``Device`` instance.
 
@@ -818,6 +882,28 @@ class Device(Registry):
         f.__belay__ = MethodMetadata(executer=ThreadExecuter, kwargs=kwargs)
         return f
 
+    def terminal(self, *, exit_char=chr(0x1D)):
+        """Start a blocking interactive terminal over the serial port."""
+        self._board.exit_raw_repl()  # In case we were previously in raw repl mode.
+        miniterm = Miniterm(self._board.serial)
+        miniterm.set_rx_encoding("UTF-8")
+        miniterm.set_tx_encoding("UTF-8")
+        miniterm.exit_character = exit_char
+        miniterm.start()
+        try:
+            miniterm.join(True)
+        except KeyboardInterrupt:
+            pass
+        miniterm.join()
+
+    def soft_reset(self):
+        """Reset device, executing ``main.py`` if available."""
+        # When in Raw REPL, ctrl-d will perform a reset, but won't execute ``main.py``
+        # https://github.com/micropython/micropython/issues/2249
+        self._board.exit_raw_repl()
+        self._board.read_until(1, b">>>")
+        self._board.ctrl_d()
+
     def _traceback_execute(
         self,
         src_file: PathType,
