@@ -11,9 +11,9 @@ class FeatureUnavailableError(BelayException):
 
 
 class SpecialFunctionNameError(BelayException):
-    """Reserved function name that may impact Belay functionality.
+    """Attempted to use a reserved Belay function name.
 
-    Currently limited to:
+    The following name rules are reserved:
 
         * Names that start and end with double underscore, ``__``.
 
@@ -27,7 +27,3 @@ class MaxHistoryLengthError(BelayException):
 
 class ConnectionLost(BelayException):
     """Lost connection to device."""
-
-
-class ConfigError(BelayException):
-    """Configuration file invalid."""
