@@ -68,10 +68,15 @@ Or:
 """
 
 import ast
+import atexit
 import itertools
 import os
+import platform
+import signal
+import subprocess
 import sys
 import time
+from threading import Lock, Thread
 
 from .webrepl import WebreplToSerial
 
@@ -82,6 +87,16 @@ except AttributeError:
     stdout = sys.stdout
 
 
+def _kill_process(pid):
+    try:
+        if platform.system() == "Windows":
+            subprocess.run(["taskkill", "/F", "/T", "/PID", str(pid)])
+        else:
+            os.killpg(os.getpgid(pid), signal.SIGTERM)
+    except ProcessLookupError:
+        pass
+
+
 def stdout_write_bytes(b):
     b = b.replace(b"\x04", b"")
     stdout.write(b)
@@ -173,38 +188,54 @@ class ProcessToSerial:
     def __init__(self, cmd):
         import subprocess
 
-        self.subp = subprocess.Popen(
+        self.subp = subp = subprocess.Popen(
             cmd,
             bufsize=0,
             shell=True,
-            preexec_fn=os.setsid,
+            start_new_session=True,
             stdin=subprocess.PIPE,
             stdout=subprocess.PIPE,
         )
+        time.sleep(0.5)
 
-        # Initially was implemented with selectors, but that adds Python3
-        # dependency. However, there can be race conditions communicating
-        # with a particular child process (like QEMU), and selectors may
-        # still work better in that case, so left inplace for now.
-        #
-        # import selectors
-        # self.sel = selectors.DefaultSelector()
-        # self.sel.register(self.subp.stdout, selectors.EVENT_READ)
+        self.buf = b""
+        self.lock = Lock()
 
-        import select
+        def process_output():
+            assert self.subp.stdout is not None
+            while True:
+                out = self.subp.stdout.read(1)
+                if out == "" and self.subp.poll() is not None:
+                    break
+                if out != "":
+                    with self.lock:
+                        self.buf += out
 
-        self.poll = select.poll()
-        self.poll.register(self.subp.stdout.fileno())
+        thread = Thread(target=process_output)
+        thread.daemon = True
+        thread.start()
+
+        sleep_multiplier = float(os.environ.get("BELAY_SLEEP_MULTIPLIER", 1.0))
+        time.sleep(5.0 * sleep_multiplier)  # Give process a chance to boot up.
+        if platform.system() == "Windows":
+            # Windows needs more time
+            time.sleep(6.0 * sleep_multiplier)
+
+        atexit.register(self.close)
 
     def close(self):
-        import signal
-
-        os.killpg(os.getpgid(self.subp.pid), signal.SIGTERM)
+        _kill_process(self.subp.pid)
+        atexit.unregister(self.close)
 
     def read(self, size=1):
-        data = b""
-        while len(data) < size:
-            data += self.subp.stdout.read(size - len(data))
+        while len(self.buf) < size:
+            # let the reading thread do its thing.
+            time.sleep(0.5)
+
+        with self.lock:
+            data = self.buf[:size]
+            self.buf = self.buf[size:]
+
         return data
 
     def write(self, data):
@@ -212,9 +243,7 @@ class ProcessToSerial:
         return len(data)
 
     def inWaiting(self):
-        # res = self.sel.select(0)
-        res = self.poll.poll(0)
-        if res:
+        if self.buf:
             return 1
         return 0
 
@@ -234,7 +263,7 @@ class ProcessPtyToTerminal:
             cmd.split(),
             bufsize=0,
             shell=False,
-            preexec_fn=os.setsid,
+            start_new_session=True,
             stdin=subprocess.PIPE,
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
@@ -330,8 +359,15 @@ class Pyboard:
 
                 time.sleep(1.0)
 
+        atexit.register(self.close)
+
     def close(self):
+        if not self.serial:
+            return
+        self.exit_raw_repl()
         self.serial.close()
+        self.serial = None
+        atexit.unregister(self.close)
 
     def read_until(self, min_num_bytes, ending, timeout=10, data_consumer=None):
         """
@@ -361,29 +397,33 @@ class Pyboard:
                 timeout_count += 1
                 if timeout is not None and timeout_count >= 100 * timeout:
                     raise PyboardError(
-                        "Timed out reading until {repr(ending)}\n    Received: {repr(data)}"
+                        f"Timed out reading until {repr(ending)}\n    Received: {repr(data)}"
                     )
                 time.sleep(0.01)
         return data
 
+    def cancel_running_program(self):
+        """Interrupts any running program."""
+        self.serial.write(b"\r\x03\x03")  # ctrl-C twice: interrupt any running program
+
+    def ctrl_d(self):
+        # Note: When in raw repl mode, a soft-reset will NOT execute ``main.py``.
+        self.serial.write(b"\x04")  # ctrl-D: soft reset
+
     def enter_raw_repl(self, soft_reset=True):
         # flush input (without relying on serial.flushInput())
         n = self.serial.inWaiting()
         while n > 0:
             self.serial.read(n)
             n = self.serial.inWaiting()
-        self.serial.write(b"\r\x03\x03")  # ctrl-C twice: interrupt any running program
+        self.cancel_running_program()
         self.exit_raw_repl()  # if device is already in raw_repl, b'>>>' won't be printed.
         self.read_until(1, b">>>")
         self.serial.write(b"\r\x01")  # ctrl-A: enter raw REPL
         if soft_reset:
             self.read_until(1, b"raw REPL; CTRL-B to exit\r\n>")
-            self.serial.write(b"\x04")  # ctrl-D: soft reset
+            self.ctrl_d()
 
-            # Waiting for "soft reboot" independently to "raw REPL" (done below)
-            # allows boot.py to print, which will show up after "soft reboot"
-            # and before "raw REPL".
-            self.read_until(1, b"soft reboot\r\n")
         self.read_until(1, b"raw REPL; CTRL-B to exit\r\n")
         self.in_raw_repl = True
 
